@@ -6,7 +6,7 @@ from typing import Dict, List, Optional, Union
 
 from fabric import Connection
 from IPython.display import HTML, display
-from novaclient.exceptions import Conflict, NotFound
+from novaclient.exceptions import NotFound
 from novaclient.v2.flavor_access import FlavorAccess as NovaFlavor
 from novaclient.v2.keypairs import Keypair as NovaKeypair
 from novaclient.v2.servers import Server as NovaServer
@@ -15,18 +15,16 @@ from packaging.version import Version
 from paramiko.client import WarningPolicy
 
 import chi
-from chi import context, util
+from chi import context, exception, util
+from chi import network as chi_network
 
 from .clients import connection, nova
-from .context import DEFAULT_IMAGE_NAME, _is_ipynb
+from .context import DEFAULT_IMAGE_NAME, _is_ipynb, session
 from .context import get as get_from_context
-from .context import session
 from .exception import CHIValueError, ResourceError, ServiceError
 from .image import Image, get_image_id, get_image_name
 from .keypair import Keypair
-from chi import network as chi_network
-from .util import random_base32, sshkey_fingerprint
-from chi import exception
+from .util import random_base32, retry_create, sshkey_fingerprint
 
 DEFAULT_IMAGE = DEFAULT_IMAGE_NAME
 DEFAULT_NETWORK = "sharednet1"
@@ -157,6 +155,8 @@ class Server:
         wait_for_active: bool = True,
         show: str = "widget",
         idempotent: bool = False,
+        retry_on_error: bool = False,
+        wait_timeout: int = 20 * 60,
         **kwargs,
     ) -> "Server":
         """
@@ -166,6 +166,8 @@ class Server:
             wait_for_active (bool, optional): Whether to wait for the server to become active before returning. Defaults to True.
             show (str, optional): The type of server information to display after creation. Defaults to "widget".
             idempotent (bool, optional): Whether to create the server only if it doesn't already exist. Defaults to False.
+            retry_on_error (bool, optional): Whether to retry the server creation if creation fails. Defaults to False.
+            wait_timeout (int): How long to wait for server to start in seconds. Default 20 minutes.
 
         Raises:
             Conflict: If the server creation fails due to a conflict and idempotent mode is not enabled.
@@ -192,18 +194,25 @@ class Server:
             net_ids=[chi_network.get_network_id(DEFAULT_NETWORK)],
             **kwargs,
         )
-        try:
-            nova_server = self.conn.compute.create_server(**server_args)
-        except Conflict as e:
-            raise ResourceError(e.message)  # Re-raise the exception if not handled
 
-        # TODO use nova_server to update self
+        def _server_create_func():
+            self.conn.compute.create_server(**server_args)
+            if wait_for_active:
+                self.wait(timeout=wait_timeout)
+            if show:
+                self.show(type=show)
 
-        if wait_for_active:
-            self.wait()
+        def _server_cleanup_func():
+            try:
+                self.delete(idempotent=True, delete_ips=False)
+                time.sleep(10)
+            except Exception:
+                # Ignore any cleanup errors
+                pass
 
-        if show:
-            self.show(type=show)
+        retry_create(
+            3 if retry_on_error else 1, _server_create_func, _server_cleanup_func
+        )
 
     @classmethod
     def _from_nova_server(cls, nova_server):
@@ -211,11 +220,7 @@ class Server:
             image_id = nova_server.image["id"]
         except Exception:
             image_id = nova_server.image_id
-
-        try:
-            flavor_id = nova_server.flavor["id"]
-        except Exception:
-            flavor_id = nova_server.flavor_id
+        flavor_name = nova_server.flavor.get("original_name", "")
 
         try:
             network_id = (
@@ -234,10 +239,12 @@ class Server:
             name=nova_server.name,
             reservation_id=None,
             image_name=get_image_name(image_id),
-            flavor_name=get_flavor(flavor_id).name,
+            flavor_name=flavor_name,
             key_name=nova_server.key_name,
             network_name=(
-                chi_network.get_network(network_id)["name"] if network_id is not None else None
+                chi_network.get_network(network_id)["name"]
+                if network_id is not None
+                else None
             ),
         )
 
@@ -277,10 +284,24 @@ class Server:
 
         return server
 
-    def delete(self) -> None:
-        """Deletes the server.
+    def delete(self, idempotent: bool = False, delete_ips: bool = True) -> None:
         """
-        delete_server(self.id)
+        Deletes the server.
+
+        Args:
+            idempotent (bool, optional): Whether to create the server only if it doesn't already exist. Defaults to False.
+            delete_ips (bool, optional): Whether to delete the server IPs from this project. Defauls to False
+        """
+        if delete_ips:
+            conn = connection(session=session())
+            for addr in self.get_all_floating_ips():
+                floating_ip_obj = chi_network.get_floating_ip(addr)
+                conn.network.delete(floating_ip_obj["id"])
+        try:
+            delete_server(self.id)
+        except NotFound:
+            if not idempotent:
+                raise ResourceError(f"Server {self.name} not found")
 
     def refresh(self):
         """
@@ -304,13 +325,16 @@ class Server:
         except Exception as e:
             raise ResourceError(f"Could not refresh server: {e}")
 
-    def wait(self, status: str = "ACTIVE", show: str = "widget") -> None:
+    def wait(
+        self, status: str = "ACTIVE", show: str = "widget", timeout: int = 20 * 60
+    ) -> None:
         """
         Waits for the server's status to reach the specified status.
 
         Args:
             status (str): The status to wait for. Defaults to "ACTIVE".
             show (str, optional): The type of server information to display after creation. Defaults to "widget".
+            timeout (int): How long to wait for server to start in seconds. Default 20 minutes.
 
         Raises:
             ServiceError: If the server does not reach the specified status within the timeout period.
@@ -319,7 +343,7 @@ class Server:
             None
         """
         print(
-            f"Waiting for server {self.name}'s status to become {status}. This typically takes 10 minutes, but can take up to 20 minutes."
+            f"Waiting for server {self.name}'s status to become {status}. This typically takes 10 minutes for baremetal, but can take up to 20 minutes."
         )
 
         pb = util.TimerProgressBar()
@@ -333,7 +357,7 @@ class Server:
                 return True
             return False
 
-        res = pb.wait(_callback, 10 * 60, 20 * 60)
+        res = pb.wait(_callback, 10 * 60, timeout)
         if not res:
             raise ServiceError(f"Timeout waiting for server to reach {status} status")
 
@@ -426,7 +450,9 @@ class Server:
                 )
         return formatted
 
-    def associate_floating_ip(self, fip: Optional[str] = None, port_id: Optional[str] = None) -> None:
+    def associate_floating_ip(
+        self, fip: Optional[str] = None, port_id: Optional[str] = None
+    ) -> None:
         """
         Associates a floating IP with the server.
 
@@ -438,20 +464,26 @@ class Server:
         Returns:
             None
         """
-        associate_floating_ip(self.id, fip, port_id)
+        fip = associate_floating_ip(self.id, fip, port_id)
         self.refresh()
+        return fip
 
-    def detach_floating_ip(self, fip: str) -> None:
+    def detach_floating_ip(self, fip: str, delete: Optional[bool] = True) -> None:
         """
         Detaches a floating IP from the server.
 
         Args:
             fip (str): The floating IP to detach.
+            delete (Optional[bool], optional): Whether to delete the floating IP after disassociation. Defaults to True.
 
         Returns:
             None
         """
         detach_floating_ip(self.id, fip)
+        if delete:
+            conn = connection(session=session())
+            floating_ip_obj = chi_network.get_floating_ip(fip)
+            conn.network.delete(floating_ip_obj["id"])
         self.refresh()
 
     def _can_connect_to_port(self, host, port, timeout):
@@ -467,11 +499,23 @@ class Server:
         Returns:
             str: Floating IP address of server
         """
+        fips = self.get_all_floating_ips()
+        if fips:
+            return fips[0]
+        return None
+
+    def get_all_floating_ips(self):
+        """Get a list of attached floating ips of this server
+
+        Returns:
+            List[str[]: Floating IP addresses of server
+        """
+        fips = []
         for net, addresses in self.addresses.items():
             for address in addresses:
                 if address.get("OS-EXT-IPS:type") == "floating":
-                    return address["addr"]
-        return None
+                    fips.append(address["addr"])
+        return fips
 
     def check_connectivity(
         self,
@@ -570,12 +614,52 @@ class Server:
         with self.ssh_connection(**kwargs) as conn:
             return conn.run(command)
 
-    def get_metadata(self):
+    def get_metadata(self) -> Dict[str, str]:
+        """Get the metadata dictionary of this server
+
+        Returns:
+            Dict[str, str]: The metadata dictionary of the server.
+        """
         return nova().servers.list_meta(self.id)["metadata"]
 
     def set_metadata_item(self, key, value):
+        """Set a metadata item for the server.
+
+        Args:
+            key (str): The metadata key
+            value (str): The metadata value
+        """
         return nova().servers.set_meta_item(self.id, key, value)
 
+    def add_security_group(self, security_group_name: str):
+        """Add a security group to the server."""
+        return nova().servers.add_security_group(self.id, security_group_name)
+
+    def remove_security_group(self, security_group_name: str):
+        """Removes a security group to the server."""
+        return nova().servers.remove_security_group(self.id, security_group_name)
+
+    def attach_volume(self, volume_id: str) -> None:
+        """Attach a Cinder volume to the server. Only supported at KVM@TACC.
+
+        Args:
+            volume_id (str): The volume to attach.
+            mount_location (str, optional): The mount location of the volume. Defaults to "/mnt/volume".
+        """
+        nova().volumes.create_server_volume(self.id, volume_id)
+
+    def detach_volume(self, volume_id: str) -> None:
+        """Detach a Cinder volume from the server. Only supported at KVM@TACC.
+
+        Args:
+            volume_id (str): The volume to detach.
+        """
+        nova().volumes.delete_server_volume(self.id, volume_id)
+
+    def __repr__(self):
+        return f"<{self.__class__.__name__} '{self.name}'>"
+
+
 ##########
 # Flavors
 ##########
@@ -587,31 +671,71 @@ class Flavor:
 
     Attributes:
         name (str): The name of the flavor.
+        description (str): The description of the flavor.
         disk (int): The disk size in GB.
         ram (int): The RAM size in MB.
         vcpus (int): The number of virtual CPUs.
+        extras (dict): Extra traits associated with this flavor.
     """
 
-    def __init__(self, name: str, disk: int, ram: int, vcpus: int):
+    def __init__(
+        self,
+        id: str,
+        name: str,
+        description: str,
+        disk: int,
+        ram: int,
+        vcpus: int,
+        extras: dict,
+    ):
+        self.id = id
         self.name = name
+        self.description = description
         self.disk = disk
         self.ram = ram
         self.vcpus = vcpus
+        self.extras = extras
 
     def __repr__(self):
-        return f"<{self.__class__.__name__} '{self.name}' (disk={self.disk}) (ram={self.ram}) (vcpus={self.vcpus})>"
+        return f"<{self.__class__.__name__} '{self.name}' {self.description} (disk={self.disk}) (ram={self.ram}) (vcpus={self.vcpus})>"
 
 
-def list_flavors() -> List[Flavor]:
+def list_flavors(reservable=None, reservation_id=None) -> List[Flavor]:
     """Get a list of all available flavors.
 
+    Args:
+        reservable (bool): Whether to filter by reservable flavors. Defaults to True.
+        reservation_id (str, optional): The reservation ID to filter by. Defaults to None.
+
     Returns:
         A list of all flavors.
     """
     if Version(context.version) >= Version("1.0"):
         nova_client = nova()
-        flavors = nova_client.flavors.list()
-        return [Flavor(name=f.name, disk=f.disk, ram=f.ram, vcpus=f.vcpus) for f in flavors]
+        flavors = nova_client.flavors.list(detailed=True)
+        chi_flavors = []
+        for f in flavors:
+            extras = f.get_keys()
+            # include a flavor if:
+            # - not filtering by reservable
+            # - is reservable in blazar & not an active reservation flavor
+            if not reservable or (
+                extras.get("aggregate_instance_extra_specs:reservation")
+                == reservation_id
+                and extras.get("trait:CUSTOM_BLAZAR_FLAVOR_RESERVATION") == "required"
+            ):
+                chi_flavors.append(
+                    Flavor(
+                        id=f.id,
+                        name=f.name,
+                        description=f.description,
+                        disk=f.disk,
+                        ram=f.ram,
+                        vcpus=f.vcpus,
+                        extras=extras,
+                    )
+                )
+        return chi_flavors
     return nova().flavors.list()
 
 
@@ -816,11 +940,13 @@ def associate_floating_ip(server_id, floating_ip_address=None, port_id=None):
     if port_id:
         port_obj = next(port for port in ports if port["id"] == port_id)
         if not port_obj:
-            raise exception.ResourceError(f"Port {port_id} not found on server {server_id}")
+            raise exception.ResourceError(
+                f"Port {port_id} not found on server {server_id}"
+            )
         ports = [port_obj]
     else:
         for port in ports:
-            floating_ip_args = {'port_id': port['id']}
+            floating_ip_args = {"port_id": port["id"]}
             try:
                 return conn.network.update_ip(
                     floating_ip_obj["id"], **floating_ip_args
@@ -828,8 +954,10 @@ def associate_floating_ip(server_id, floating_ip_address=None, port_id=None):
             except SDKException:
                 # Ignore errors and try the next port
                 pass
-    floating_ip_address = floating_ip_obj["floating_ip_obj"]
-    raise exception.ResourceError(f"None of the ports can route to floating ip {floating_ip_address} on server {server_id}")
+    floating_ip_address = floating_ip_obj["floating_ip_address"]
+    raise exception.ResourceError(
+        f"None of the ports can route to floating ip {floating_ip_address} on server {server_id}"
+    )
 
 
 def detach_floating_ip(server_id, floating_ip_address):
@@ -952,8 +1080,11 @@ def get_keypair(name=None) -> Keypair:
         name = get_from_context("keypair_name")
 
     nova_client = nova()
-    keypair = nova_client.keypairs.get(name)
-    return Keypair(name=keypair.name, public_key=keypair.public_key)
+    try:
+        keypair = nova_client.keypairs.get(name)
+        return Keypair(name=keypair.name, public_key=keypair.public_key)
+    except NotFound:
+        return Keypair(name=name, public_key=None)
 
 
 def list_keypair() -> List[Keypair]:
